@@ -21,10 +21,12 @@
 
 #include "config.h"
 
+#include <netdb.h>
+#include <string.h>
 #include <polkit/polkit.h>
 #include "polkitbackendlocalauthorizationstore.h"
 
-/**
+/* <internal>
  * SECTION:polkitbackendlocalauthorizationstore
  * @title: PolkitBackendLocalAuthorizationStore
  * @short_description: Watches a directory for authorization files
@@ -73,12 +75,19 @@ typedef struct
 {
   gchar *id;
 
+  /* Identities with glob support */
   GList *identity_specs;
+
+  /* Netgroup identity strings, which can not support glob syntax */
+  GList *netgroup_identities;
+
   GList *action_specs;
 
   PolkitImplicitAuthorization result_any;
   PolkitImplicitAuthorization result_inactive;
   PolkitImplicitAuthorization result_active;
+
+  GHashTable *return_value;
 } LocalAuthorization;
 
 static void
@@ -87,8 +96,11 @@ local_authorization_free (LocalAuthorization *authorization)
   g_free (authorization->id);
   g_list_foreach (authorization->identity_specs, (GFunc) g_pattern_spec_free, NULL);
   g_list_free (authorization->identity_specs);
+  g_list_free_full (authorization->netgroup_identities, g_free);
   g_list_foreach (authorization->action_specs, (GFunc) g_pattern_spec_free, NULL);
   g_list_free (authorization->action_specs);
+  if (authorization->return_value != NULL)
+    g_hash_table_unref (authorization->return_value);
   g_free (authorization);
 }
 
@@ -105,6 +117,7 @@ local_authorization_new (GKeyFile      *key_file,
   gchar *result_any_string;
   gchar *result_inactive_string;
   gchar *result_active_string;
+  gchar **return_value_strings;
   guint n;
 
   identity_strings = NULL;
@@ -112,6 +125,7 @@ local_authorization_new (GKeyFile      *key_file,
   result_any_string = NULL;
   result_inactive_string = NULL;
   result_active_string = NULL;
+  return_value_strings = NULL;
 
   authorization = g_new0 (LocalAuthorization, 1);
 
@@ -128,8 +142,13 @@ local_authorization_new (GKeyFile      *key_file,
     }
   for (n = 0; identity_strings[n] != NULL; n++)
     {
-      authorization->identity_specs = g_list_prepend (authorization->identity_specs,
-                                                      g_pattern_spec_new (identity_strings[n]));
+      /* Put netgroup entries in a seperate list from other identities who support glob syntax */
+      if (g_str_has_prefix (identity_strings[n], "unix-netgroup:"))
+        authorization->netgroup_identities = g_list_prepend (authorization->netgroup_identities,
+                                                             g_strdup (identity_strings[n] + sizeof "unix-netgroup:" - 1));
+      else
+        authorization->identity_specs = g_list_prepend (authorization->identity_specs,
+                                                        g_pattern_spec_new (identity_strings[n]));
     }
 
   action_strings = g_key_file_get_string_list (key_file,
@@ -221,6 +240,42 @@ local_authorization_new (GKeyFile      *key_file,
           goto out;
     }
 
+  return_value_strings = g_key_file_get_string_list (key_file,
+                                                     group,
+                                                     "ReturnValue",
+                                                     NULL,
+                                                     error);
+  if (return_value_strings != NULL)
+    {
+      for (n = 0; return_value_strings[n] != NULL; n++)
+        {
+          gchar *p;
+          const gchar *key;
+          const gchar *value;
+
+          p = strchr (return_value_strings[n], '=');
+          if (p == NULL)
+            {
+              g_warning ("Item `%s' in ReturnValue is malformed. Ignoring.",
+                         return_value_strings[n]);
+              continue;
+            }
+
+          *p = '\0';
+          key = return_value_strings[n];
+          value = p + 1;
+
+          if (authorization->return_value == NULL)
+            {
+              authorization->return_value = g_hash_table_new_full (g_str_hash,
+                                                                   g_str_equal,
+                                                                   g_free,
+                                                                   g_free);
+            }
+          g_hash_table_insert (authorization->return_value, g_strdup (key), g_strdup (value));
+        }
+    }
+
   authorization->id = g_strdup_printf ("%s::%s", filename, group);
 
  out:
@@ -229,6 +284,7 @@ local_authorization_new (GKeyFile      *key_file,
   g_free (result_any_string);
   g_free (result_inactive_string);
   g_free (result_active_string);
+  g_strfreev (return_value_strings);
   return authorization;
 }
 
@@ -474,6 +530,12 @@ polkit_backend_local_authorization_store_new (GFile       *directory,
 static void
 polkit_backend_local_authorization_store_purge (PolkitBackendLocalAuthorizationStore *store)
 {
+  gchar *path;
+
+  path = g_file_get_path (store->priv->directory);
+  g_debug ("Dropping all .pkla caches for directory `%s'", path);
+  g_free (path);
+
   g_list_foreach (store->priv->authorizations, (GFunc) local_authorization_free, NULL);
   g_list_free (store->priv->authorizations);
   store->priv->authorizations = NULL;
@@ -499,7 +561,7 @@ polkit_backend_local_authorization_store_ensure (PolkitBackendLocalAuthorization
 
   error = NULL;
   enumerator = g_file_enumerate_children (store->priv->directory,
-                                          "standard::*",
+                                          "standard::name",
                                           G_FILE_QUERY_INFO_NONE,
                                           NULL,
                                           &error);
@@ -591,6 +653,8 @@ polkit_backend_local_authorization_store_ensure (PolkitBackendLocalAuthorization
       g_free (filename);
     }
 
+  store->priv->has_data = TRUE;
+
  out:
   g_list_foreach (files, (GFunc) g_object_unref, NULL);
   g_list_free (files);
@@ -605,6 +669,7 @@ polkit_backend_local_authorization_store_ensure (PolkitBackendLocalAuthorization
  * @out_result_any: Return location for the result for any subjects if the look up matched.
  * @out_result_inactive: Return location for the result for subjects in local inactive sessions if the look up matched.
  * @out_result_active: Return location for the result for subjects in local active sessions if the look up matched.
+ * @out_details: %NULL or a #PolkitDetails object to append key/value pairs to on a positive match.
  *
  * Checks if an authorization entry from @store matches @identity, @action_id and @details.
  *
@@ -618,7 +683,8 @@ polkit_backend_local_authorization_store_lookup (PolkitBackendLocalAuthorization
                                                  PolkitDetails                        *details,
                                                  PolkitImplicitAuthorization          *out_result_any,
                                                  PolkitImplicitAuthorization          *out_result_inactive,
-                                                 PolkitImplicitAuthorization          *out_result_active)
+                                                 PolkitImplicitAuthorization          *out_result_active,
+                                                 PolkitDetails                        *out_details)
 {
   GList *l, *ll;
   gboolean ret;
@@ -650,7 +716,7 @@ polkit_backend_local_authorization_store_lookup (PolkitBackendLocalAuthorization
       if (ll == NULL)
         continue;
 
-      /* then match the identity */
+      /* then match the identity against identity specs */
       if (identity_string == NULL)
         identity_string = polkit_identity_to_string (identity);
       for (ll = authorization->identity_specs; ll != NULL; ll = ll->next)
@@ -658,6 +724,22 @@ polkit_backend_local_authorization_store_lookup (PolkitBackendLocalAuthorization
           if (g_pattern_match_string ((GPatternSpec *) ll->data, identity_string))
             break;
         }
+
+      /* if no identity specs matched and identity is a user, match against netgroups */
+      if (ll == NULL && POLKIT_IS_UNIX_USER (identity))
+        {
+          PolkitUnixUser *user_identity = POLKIT_UNIX_USER (identity);
+          const gchar *user_name = polkit_unix_user_get_name (user_identity);
+          if (!user_name)
+            continue;
+
+          for (ll = authorization->netgroup_identities; ll != NULL; ll = ll->next)
+            {
+              if (innetgr ((const gchar *) ll->data, NULL, user_name, NULL))
+                break;
+            }
+        }
+
       if (ll == NULL)
         continue;
 
@@ -667,6 +749,19 @@ polkit_backend_local_authorization_store_lookup (PolkitBackendLocalAuthorization
       *out_result_active = authorization->result_active;
       ret = TRUE;
 
+      if (out_details != NULL && authorization->return_value != NULL)
+        {
+          GHashTableIter iter;
+          const gchar *key;
+          const gchar *value;
+
+          g_hash_table_iter_init (&iter, authorization->return_value);
+          while (g_hash_table_iter_next (&iter, (gpointer *) &key, (gpointer *) &value))
+            {
+              polkit_details_insert (out_details, key, value);
+            }
+        }
+
 #if 0
       g_debug ("authorization with id `%s' matched action_id `%s' for identity `%s'",
                authorization->id,
